import logging
from durin.actuator import StreamOn

from durin.actuator import DurinActuator
from durin.io.network import TCPLink, UDPLink
from durin.sensor import DurinSensor
import durin.io

DURIN_CONTROLLER_PORT_TCP = 1337
DURIN_CONTROLLER_PORT_UDP = 4305

DURIN_DVS_PORT_TCP = 2301


class Durin:
    """
    Interface to the Durin robot.
    This class first establishes connection to the controller and DVS system on board the robot.
    Afterwards, commands can be sent to the robot and sensory data can be `.read` from the robot.

    Note that the sensor observation is described in the sensor.Observation class.
    The DVS output is a 640x480 PyTorch tensor (if the aestream dependency is installed).

    Example:

    >>> with Durin(10.0.0.1) as durin:
    >>>   while True:
    >>>     observation, dvs = durin.read()
    >>>     ...
    >>>     durin(Move(x, y, rot))

    Arguments:
        durin_ip (str): The IPv4 address of the Durin microcontroller
        device (str): The PyTorch device to use for storing tensors. Defaults to cpu
        stream_command (Optional[StreamOn]): The command sent to the Durin microcontroller upon start. Can be customized, but uses sensible values by default.
        sensor_frequency (int): The update frequency of the sensor information in ms. Defaults to 15ms.
        disable_dvs (bool): Disables connection to DVS. Useful if necessary libraries are not installed. Defaults to False.
    """

    def __init__(
        self,
        durin_ip: str,
        device: str = "cpu",
        stream_command: Optional[StreamOn] = None,
        sensor_frequency: int = 15,
        disable_dvs: bool = True,
    ):
        if stream_command is not None:
            self.stream_command = stream_command
        else:
            response_ip = durin.io.network.get_ip(durin_ip)
            self.stream_command = StreamOn(
                response_ip, DURIN_CONTROLLER_PORT_UDP, sensor_frequency
            )

        # Controller
        tcp_link = TCPLink(durin_ip, DURIN_CONTROLLER_PORT_TCP)
        udp_link = UDPLink(self.stream_command.host, self.stream_command.port)
        self.sensor = DurinSensor(udp_link)
        self.actuator = DurinActuator(tcp_link)

        # DVS
        self.disable_dvs = disable_dvs
        if not disable_dvs:
            import dvs

            ip_list = durin_ip.split(".")
            dvs_ip = ".".join(ip_list[:-1]) + f".{int(ip_list[-1]) + 10}"
            self.dvs_client = dvs.DVSClient(dvs_ip, DURIN_DVS_PORT_TCP)
            self.dvs = dvs.DVSSensor((640, 480), device, self.stream_command.port + 1)
        else:
            logging.debug("DVS output disabled")

    def __enter__(self):
        # Controller
        self.sensor.start()
        self.actuator.start()
        # Start streaming
        self(self.stream_command)

        logging.debug(
            f"Durin Controller receiving on {self.stream_command.host}:{self.stream_command.port}"
        )

        # DVS
        if not self.disable_dvs:
            self.dvs.start_stream()
            logging.debug(f"Durin DVS sending to {self.dvs_client.address}")
            self.dvs_client.start_stream(
                self.stream_command.host, self.stream_command.port + 1
            )
            logging.debug(
                f"Durin DVS receiving on {self.stream_command.host}:{self.stream_command.port + 1}"
            )

        return self

    def __exit__(self, e, b, t):
        self.sensor.stop()
        self.actuator.stop()
        if not self.disable_dvs:
            self.dvs_client.stop_stream()
            self.dvs.stop_stream()

    def __call__(self, command):
        return self.actuator(command)

    def update_frequency(self) -> float:
        return 1 / self.sensor.freq / 6

    def read(self):
        """
        Retrieves sensor data from Durin

        Returns:
            Tuple of Observation, DVS tensors, and Command reply
        """
        durin = self.sensor.read()
        if not self.disable_dvs:
            dvs = self.dvs.read()
        else:
            dvs = None
        cmd = self.actuator.read()
        return (durin, dvs, cmd)



from abc import ABC, abstractmethod
import multiprocessing
import time
from typing import Generic, NamedTuple, Tuple, TypeVar

import numpy as np
from durin.io import SENSORS

from durin.io.network import UDPLink
from durin.io.runnable import RunnableConsumer
from durin.io.ringbuffer import RingBuffer

T = TypeVar("T")


class Observation(NamedTuple):
    tof: np.ndarray = np.zeros((8, 8, 8))
    charge: float = 0
    voltage: float = 0
    imu: np.ndarray = np.zeros((3, 3))
    uwb: float = 0
    frequency: float = 0

    def __repr__(self) -> str:
        tof_str = " ".join([f"{x.mean():.0f}±{x.std():.0f}" for x in self.tof])
        return f"Durin {self.battery}%\n\tIMU: {self.imu}\n\tTOF: {tof_str}"


class Sensor(ABC, Generic[T]):
    @abstractmethod
    def read(self) -> T:
        pass


class DurinSensor(RunnableConsumer, Sensor[Observation]):
    def __init__(self, link: UDPLink):
        self.link = link
        context = multiprocessing.get_context("spawn")
        self.tof = context.Array("f", 8 * 8 * 8)
        self.charge = context.Value("f", 0)
        self.voltage = context.Value("f", 0)
        self.imu = context.Array("d", 3 * 3)
        self.uwb = context.Value("d", 0)
        self.ringbuffer = context.Array("d", 50)
        self.ringbuffer_idx = context.Value("i", 0)
        self.timestamp_update = context.Value("d", time.time())

        super().__init__(
            self.link.buffer,
            self.tof,
            self.charge,
            self.voltage,
            self.imu,
            self.uwb,
            self.ringbuffer,
            self.ringbuffer_idx,
            self.timestamp_update,
        )

    def read(self) -> Observation:
        tof = np.frombuffer(self.tof.get_obj(), dtype=np.float32).reshape((8, 8, 8))
        imu = np.frombuffer(self.imu.get_obj()).reshape((3, 3))
        frequency = 1 / (np.frombuffer(self.ringbuffer.get_obj()).mean() + 1e-7)
        return Observation(
            tof,
            charge=self.charge.value,
            voltage=self.voltage.value,
            imu=imu,
            uwb=self.uwb.value,
            frequency=frequency,
        )

    def start(self):
        super().start()
        self.link.start()

    def stop(self):
        super().stop()
        self.link.stop()

    def consume(
        self,
        item,
        tof,
        charge,
        voltage,
        imu,
        uwb,
        ringbuffer,
        ringbuffer_idx,
        timestamp_update,
    ):
        which = item.which()
        if which == "tofObservations":
            for obs in item.tofObservations.observations:
                data = np.array(obs.ranges)
                tof.get_obj()[obs.id * 64: (obs.id + 1) * 64] = data
        elif which == "systemStatus":
            charge.value = item.systemStatus.batteryPercent
            voltage.value = item.systemStatus.batteryMv
        # TODO: Add more sensors

        # Update Hz
        time_now = time.time()
        buffer = RingBuffer(np.frombuffer(ringbuffer.get_obj()))
        buffer.counter = ringbuffer_idx.value
        buffer.append(time_now - timestamp_update.value)
        ringbuffer.get_obj()[:] = buffer.buffer
        ringbuffer_idx.value = buffer.counter
        timestamp_update.value = time_now



import math
import sdl2.ext

import numpy as np
from durin.actuator import Move

from durin.durin import Durin
from durin.io.gamepad import Gamepad

TOF_WIDTH = 8  # 8*8
TOF_HEIGHT = 8

DVS_HEIGHT = 480
DVS_WIDTH = 640

import numpy as np

SENSOR_PLACEMENTS = [
    (0.97, 0.5),
    (0.97, 0.03),
    (0.5, 0.03),
    (0.97, 0.97),
    (0.5, 0.97),
    (0.02, 0.97),
    (0.03, 0.03),
    (0.02, 0.5),
]

MAX_WHEEL = 600


def to_rgba(value):
    alpha = 255
    red = green = blue = value
    return (alpha << 24) | (red << 16) | (green << 8) | blue


def tof_sensor_to_pixels(matrix, size: int = 128):
    pixels = np.empty((size, size))
    width = math.ceil(size / len(matrix))
    for idr, row in enumerate(matrix):
        for idx, element in enumerate(row):
            pixels[
                idr * width : (idr + 1) * width, idx * width : (idx + 1) * width
            ] = to_rgba(element)
    return pixels.T


class DurinUI(Durin):
    """
    **Graphical** Interface to the Durin robot (as opposed to the native durin.Durin class).
    Note that the evaluation may be slower than the native class, due to graphics and user I/O overheads.

    Example:

    >>> with DurinUI(10.0.0.1) as durin:
    >>>   while True:
    >>>     data, dvs, cmd = durin.read()
    >>>     durin.render_sensors(data)
    >>>     durin.read_user_input()
    >>>     durin(Move(x, y, rot))

    Arguments:
        durin_ip (str): The IPv4 address of the Durin microcontroller
        device (str): The PyTorch device to use for storing tensors. Defaults to cpu
        stream_command (Optional[StreamOn]): The command sent to the Durin microcontroller upon start. Can be customized, but uses sensible values by default.
        sensor_frequency (int): The update frequency of the sensor information in ms. Defaults to 15ms.
        disable_dvs (bool): Disables connection to DVS. Useful if necessary libraries are not installed. Defaults to False.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.resources = sdl2.ext.Resources(__file__)
        self.gamepad = Gamepad()

        self.vertical = 0
        self.horizontal = 0
        self.tau = 0.9999
        self.rot = 0

    def __enter__(self):
        sdl2.ext.init()

        self.window = sdl2.ext.Window("Durin Dashboard", size=(1000, 900))
        self.window.show()

        factory = sdl2.ext.SpriteFactory(sdl2.ext.SOFTWARE)
        bg = factory.from_image(self.resources.get_path("durin_birdseye.jpg"))
        renderer = factory.create_sprite_render_system(self.window)
        renderer.render(bg)

        self.pixels = sdl2.ext.pixelaccess.pixels2d(renderer)
        self.matrix = np.random.randint(255, size=(8, 8))

        self.gamepad.start()
        return super().__enter__()

    def __exit__(self, e, b, t):
        self.window.close()
        self.gamepad.stop()
        return super().__exit__(e, b, t)

    def read_user_input(self, allow_movement: bool = True):
        events = sdl2.ext.get_events()

        # Gamepad
        if not self.gamepad.queue.empty():
            x, y, r = self.gamepad.queue.get()
            self.horizontal = x
            self.vertical = y
            self.rot = r

        # Keyboard
        for event in events:
            if event.type == sdl2.SDL_QUIT or (
                event.type == sdl2.SDL_KEYDOWN
                and event.key.keysym.sym == sdl2.SDLK_ESCAPE
            ):
                return False
            elif event.type == sdl2.SDL_KEYDOWN:
                if (
                    event.key.keysym.sym == sdl2.SDLK_UP
                    or event.key.keysym.sym == sdl2.SDLK_w
                ):
                    self.vertical = 500
                elif (
                    event.key.keysym.sym == sdl2.SDLK_LEFT
                    or event.key.keysym.sym == sdl2.SDLK_a
                ):  # Left
                    self.horizontal = -500
                elif (
                    event.key.keysym.sym == sdl2.SDLK_RIGHT
                    or event.key.keysym.sym == sdl2.SDLK_d
                ):  # Right
                    self.horizontal = 500
                elif (
                    event.key.keysym.sym == sdl2.SDLK_DOWN
                    or event.key.keysym.sym == sdl2.SDLK_s
                ):
                    self.vertical = -500

                if event.key.keysym.sym == sdl2.SDLK_BACKSPACE:
                    self.horizontal = 0
                    self.vertical = 0
            elif event.type == sdl2.SDL_KEYUP:
                if (
                    event.key.keysym.sym == sdl2.SDLK_UP
                    or event.key.keysym.sym == sdl2.SDLK_w
                    or event.key.keysym.sym == sdl2.SDLK_DOWN
                    or event.key.keysym.sym == sdl2.SDLK_s
                ):
                    self.vertical = 0
                elif (
                    event.key.keysym.sym == sdl2.SDLK_LEFT
                    or event.key.keysym.sym == sdl2.SDLK_a
                    or event.key.keysym.sym == sdl2.SDLK_RIGHT
                    or event.key.keysym.sym == sdl2.SDLK_d
                ):
                    self.horizontal = 0
            elif event.type == sdl2.SDL_MOUSEMOTION:
                self.rot = int(self.rot + min(10, max(-10, event.motion.xrel)) * 2)

        if allow_movement:
            self(Move(self.horizontal, self.vertical, self.rot))
        return True

    def render_sensors(self, obs, size: int = 180):
        tofs = (np.tanh((obs.tof / 500)) * 255).astype(np.int32)
        width, height = self.pixels.shape
        for idt, tof in enumerate(tofs):
            xr, yr = SENSOR_PLACEMENTS[idt]
            x = int((xr * (width - size)))
            y = int((yr * (height - size)))
            self.pixels[x : x + size, y : y + size] = tof_sensor_to_pixels(tof, size)

        self.window.refresh()
